@@ -3,6 +3,8 @@
 
 from __future__ import print_function
 
+import contextlib
+import io
 import json
 import logging
 import os
@@ -24,6 +26,7 @@ from IPython.core.magic import (
 )
 from IPython.core.magic_arguments import magic_arguments, argument, parse_argstring
 import sparkmagic.utils.configuration as conf
+from ..utils.resource_check import check_host_and_port
 
 # The class MUST call this class decorator at creation time
 # https://ipython.readthedocs.io/en/stable/config/custommagics.html
@@ -56,14 +59,12 @@ logger = logging.getLogger(__name__)
 logger.setLevel(logging.INFO)
 logger.addHandler(logging.StreamHandler(sys.stdout))
 
-"""
-This method handles connectivity to EMR clusters from PySpark and SparkScala kernels
-"""
-
-
 def _magic_kernel_connect_to_emr(
     emr_client, args, emr_cluster, username, password, kernel_name
 ):
+    """
+    Handles connectivity to EMR clusters from PySpark and SparkScala kernels
+    """
     print("Initiating EMR connection..")
     ipy = get_ipython()
     _run_preset_cell_magics(ipy)
@@ -80,15 +81,16 @@ def _magic_kernel_connect_to_emr(
 
     # Start spark session
     start_session_magic = ipy.find_cell_magic("_do_not_call_start_session")
-    start_session_magic("")
+    error_message = None
+    session_started = start_session_magic("")
+    if not session_started:
+        error_message = "Failed to start spark session."
 
-    # TODO: Find a way to establish actual success of spark session establishment, before assigning success code and
-    #  error message, across kernels
     _echo_response_to_iopub_web_socket(
         _build_response(
             cluster_id=args.cluster_id,
-            error_message=None,
-            success=True,
+            error_message=error_message,
+            success=session_started,
             service="emr",
             operation="connect",
         )
@@ -103,6 +105,10 @@ def _get_endpoint_magic_line(
         args=args, emr_client=emr_client, emr_cluster=emr_cluster
     )
 
+    # fast fail by check livy port availability. If fail, do not proceed to compose the endpoint.
+    if not check_host_and_port(livy_host, int(livy_port)):
+        raise RuntimeError(f"Cannot connect to livy service: {livy_host}:{livy_port}")
+
     livy_endpoint = "http://{0}:{1}".format(livy_host, livy_port)
     if kernel_name in MAGIC_KERNELS:
         return _get_magic_kernel_endpoint_magic_line(
@@ -170,71 +176,83 @@ def _get_lock():
             lock.release(force=True)
     return lock
 
+class IpythonWidgetBasedAuthProvider:
+    """
+    Class responsible for retrieving credentials needed to connect to EMR cluster.
+    """
+    def get_credentials_and_connect_to_cluster(self, emr_client, args, emr_cluster, kernel_name):
+        username = widgets.Text(
+            value="", placeholder="", description="Username:", disabled=False
+        )
+        password = widgets.Password(
+            value="", placeholder="Enter password", description="Password:", disabled=False
+        )
+        display(username, password)
+        button = widgets.Button(
+            description="Connect",
+            disabled=False,
+            button_style="",
+            tooltip="Connect to EMR",
+            icon="",
+        )
+        output = widgets.Output()
+        display(button, output)
 
-def _get_credentials_and_connect(emr_client, args, emr_cluster, kernel_name):
-    username = widgets.Text(
-        value="", placeholder="", description="Username:", disabled=False
-    )
-    password = widgets.Password(
-        value="", placeholder="Enter password", description="Password:", disabled=False
-    )
-    display(username, password)
-    button = widgets.Button(
-        description="Connect",
-        disabled=False,
-        button_style="",
-        tooltip="Connect to EMR",
-        icon="",
-    )
-    output = widgets.Output()
-    display(button, output)
+        # Do not remove the unused param otherwise the button will become unresponsive
+        def _on_button_clicked(b):
+            with output:
+                _connect_to_cluster(
+                    emr_client=emr_client,
+                    args=args,
+                    emr_cluster=emr_cluster,
+                    kernel_name=kernel_name,
+                    username=username.value,
+                    password=password.value
+                )
 
-    # Do not remove the unused param otherwise the button will become unresponsive
-    def _on_button_clicked(b):
-        with output:
-            if args.auth_type == "Basic_Access":
+        button.on_click(_on_button_clicked)
+
+def _connect_to_cluster(emr_client, args, emr_cluster, kernel_name, username, password):
+    if args.auth_type == "Basic_Access":
+        _initiate_connect_based_on_kernel(
+            emr_client=emr_client,
+            args=args,
+            emr_cluster=emr_cluster,
+            username=username,
+            password=password,
+            kernel_name=kernel_name,
+        )
+    else:
+        # At this point lock should be available to acquire if another process has not already taken it
+        # Go ahead and try to use the "with" context directly
+        with _get_lock().acquire(
+                # Try for 15 seconds to acquire the lock, otherwise fail with lock acquire failure
+                timeout=15
+        ):
+            completed_process = _generate_kerberos_token(
+                emr_cluster=emr_cluster, username=username, password=password
+            )
+            if completed_process.returncode == 0:
                 _initiate_connect_based_on_kernel(
                     emr_client=emr_client,
                     args=args,
                     emr_cluster=emr_cluster,
-                    username=username.value,
-                    password=password.value,
+                    username=None,
+                    password=None,
                     kernel_name=kernel_name,
                 )
             else:
-                # At this point lock should be available to acquire if another process has not already taken it
-                # Go ahead and try to use the "with" context directly
-                with _get_lock().acquire(
-                    # Try for 15 seconds to acquire the lock, otherwise fail with lock acquire failure
-                    timeout=15
-                ):
-                    completed_process = _generate_kerberos_token(
-                        emr_cluster=emr_cluster, username=username, password=password
-                    )
-                    if completed_process.returncode == 0:
-                        _initiate_connect_based_on_kernel(
-                            emr_client=emr_client,
-                            args=args,
-                            emr_cluster=emr_cluster,
-                            username=None,
-                            password=None,
-                            kernel_name=kernel_name,
-                        )
-                    else:
-                        _handle_kerberos_connectivity_failure(
-                            args=args, completed_process=completed_process
-                        )
-
-    button.on_click(_on_button_clicked)
-
+                _handle_kerberos_connectivity_failure(
+                    args=args, completed_process=completed_process
+                )
 
 def _generate_kerberos_token(emr_cluster, username, password):
     # noinspection PyTypeChecker
     write_krb_conf(emr_cluster, KRB_FILE_PATH)
-    cmd = ["kinit", username.value.encode()]
+    cmd = ["kinit", username.encode()]
     completed_process = subprocess.run(
         cmd,
-        input=password.value.encode(),
+        input=password.encode(),
         stdout=subprocess.PIPE,
         stderr=subprocess.PIPE,
     )
@@ -342,11 +360,22 @@ def _ipython_kernel_connect_to_emr(
     ipy.run_line_magic("spark", "cleanup")
     ipy.run_line_magic("spark", endpoint_magic_line)
 
+    # Check if session is created. Return true if sagemaker session name prefix is in '%spark info' response.
+    # Sagemaker session name prefix is under control and can be used as stable contract.
+    spark_info_output = io.StringIO()
+    with contextlib.redirect_stdout(spark_info_output):
+        ipy.run_line_magic("spark", "info")
+
+    session_started = SPARK_SESSION_NAME_PREFIX in spark_info_output.getvalue()
+    error_message = None
+    if not session_started:
+        error_message = "Failed to start spark session."
+
     _echo_response_to_iopub_web_socket(
         _build_response(
             cluster_id=args.cluster_id,
-            error_message=None,
-            success=True,
+            error_message=error_message,
+            success=session_started,
             service="emr",
             operation="connect",
         )
@@ -419,7 +448,8 @@ def _initiate_emr_connect(args, boto_session, kernel_name):
     )
 
     if cluster.is_krb_cluster or args.auth_type == "Basic_Access":
-        _get_credentials_and_connect(
+        authProvider = IpythonWidgetBasedAuthProvider()
+        authProvider.get_credentials_and_connect_to_cluster(
             emr_client=emr_client,
             args=args,
             emr_cluster=cluster,
